!#################################################################
!
! Fortran module around UDUnits libraries (version 1 and 2) .
!
! VERSION WRAPPER
!
!   This module is a wrapper to either udunits1 or UDUnits.
!   Which interface is selected depends on macro definitions:
!
!     with_udunits      : version 1
!     with_udunits2     : version 2
!
!   The macro definition should be placed in the 
!   macro include file "udunits_version.inc":
!
!     #define with_udunits2
!
!   If none of these macro's is defined, the code can still be compiled,
!   but every call to routine will result in an error status.
!
!
! USAGE
!
!   use UDUnits, only : UDUnits_NOERR
!   use UDUnits, only : UDUnits_StrError
!   use UDUnits, only : UDUnits_Init
!   use UDUnits, only : UDUnits_Done
!
!   integer                     ::  status
!   character(len=64)           ::  spec, spec2
!   real(8)                     ::  factor
!
!   ! * module initialisation
!
!   call UDUnits_Init( status )
!   if (status/=UDUnits_NOERR) then
!     print *, trim(UDUnits_StrError(status))
!     stop
!   end if
!
!   ! * high levell routines
!  
!   spec = 'kg/s'
!   call UDUnits_Standard( spec, spec2, status )
!   write (*,'("standard name of `",a,"` is `",a,"`")') trim(spec), trim(spec2)
!  
!   spec = 'gram/cm3' ; spec2 = 'kg/m3'
!   call UDUnits_ConversionFactor( spec, spec2, slope, status )
!   write (*,'("conversion factor from `",a,"` to `",a,"` is ",f12.4)') trim(spec), trim(spec2), slope
!
!   ! * done with module
!
!   call UDUnits_Done( status )
!
!
! CONTENT
!
!   The module consists of the following files:
!
!     udunits.F90          : main module
!     udunits_version.inc  : macro include file for version selection
!
!       udunits1.F90         : high level interface to udunits v1 library
!         udunits1_inc.F       : interface to library include file
!
!       udunits2.F90         : high level interface to udunits v2 library
!         udunits2_f.f90       : fortran-to-c wrapper for udunits2
!         udunits2_f.inc
!
!
! HISTORY
!   2015-12, Arjo Segers, TNO
!
!
!### macro's #####################################################
!
#define TRACEBACK write (*,'("in ",a," (",a,", line",i5,")")') rname, __FILE__, __LINE__
!
#define IF_NOT_OK_RETURN(action) if (status/=0) then; action; return; end if
!
#include "udunits_version.inc"
!
!#################################################################


module UDUnits

  implicit none
  

  ! --- in/out -----------------------------------
  
  private
  
  public    ::  UDUNITS_NOERR, UDUnits_StrError

  public    ::  UDUnits_Init, UDUnits_Done
  
  public    ::  UDUnits_Standard
  public    ::  UDUnits_ConversionFactor


  ! --- const --------------------------------

  ! module name:
  character(len=*), parameter  ::  mname = 'UDUnits'

  ! no error:
  integer, parameter  ::  UDUNITS_NOERR = 0
  

  ! --- var --------------------------------

  ! storage for latest error:
  character(len=256)  ::  error_message = ''
  

contains


  ! ====================================================================
  ! ===
  ! === module routines
  ! ===
  ! ====================================================================
  
  
  subroutine UDUnits_Init( status )

#ifdef with_udunits
    use UDUnits1, only : UDUnits1_Init
#else
#ifdef with_udunits2
    use UDUnits2, only : UDUnits2_Init
#endif
#endif
  
    ! --- in/out ---------------------------------
    
    integer, intent(out)      ::  status
    
    ! --- const ----------------------------------

    character(len=*), parameter ::  rname = mname//'/UDUnits_Init'
    
    ! --- local ----------------------------------
    
    ! --- begin ----------------------------------

#ifdef with_udunits
    call UDUnits1_Init( status )
    IF_NOT_OK_RETURN(status=1)
#else
#ifdef with_udunits2
    call UDUnits2_Init( status )
    IF_NOT_OK_RETURN(status=1)
#else
    write (error_message,'("UDUnits requires definition of either with_udunits or with_udunits2 macro")')
    status=1; return
#endif
#endif
    
    ! ok
    status = 0
    
  end subroutine UDUnits_Init
  
  
  ! ***


  subroutine UDUnits_Done( status )

#ifdef with_udunits
    use UDUnits1, only : UDUnits1_Done
#else
#ifdef with_udunits2
    use UDUnits2, only : UDUnits2_Done
#endif
#endif
  
    ! --- in/out ---------------------------------
    
    integer, intent(out)      ::  status
    
    ! --- const ----------------------------------

    character(len=*), parameter ::  rname = mname//'/UDUnits_Done'
    
    ! --- begin ----------------------------------

#ifdef with_udunits
    call UDUnits1_Done( status )
    IF_NOT_OK_RETURN(status=1)
#else
#ifdef with_udunits2
    call UDUnits2_Done( status )
    IF_NOT_OK_RETURN(status=1)
#else
    write (error_message,'("UDUnits requires definition of either with_udunits or with_udunits2 macro")')
    status=1; return
#endif
#endif
    
    ! ok
    status = 0
    
  end subroutine UDUnits_Done
  
  
  ! ====================================================================
  ! ===
  ! === error messages
  ! ===
  ! ====================================================================
  

  function UDUnits_StrError( status )

#ifdef with_udunits
    use UDUnits1, only : UDUnits1_StrError
#else
#ifdef with_udunits2
    use UDUnits2, only : UDUnits2_StrError
#endif
#endif

    ! --- in/out ---------------------------------
    
    character(len=256)          ::  UDUnits_StrError
    integer, intent(inout)      ::  status
    
    ! --- const ----------------------------------

    character(len=*), parameter ::  rname = mname//'/UDUnits_StrError'
    
    ! --- begin ----------------------------------

#ifdef with_udunits
    UDUnits_StrError = trim(UDUnits1_StrError(status))
#else
#ifdef with_udunits2
    UDUnits_StrError = trim(UDUnits2_StrError(status))
#else
    write (error_message,'("UDUnits requires definition of either with_udunits or with_udunits2 macro")')
    UDUnits_StrError = trim(error_message)
    status=1; return
#endif
#endif
    
  end function UDUnits_StrError
  

  ! ====================================================================
  ! ===
  ! === high level routines
  ! ===
  ! ====================================================================
  
  
  subroutine UDUnits_Standard( spec_from, spec_to, status )

#ifdef with_udunits
    use UDUnits1, only : UDUnits1_Standard
#else
#ifdef with_udunits2
    use UDUnits2, only : UDUnits2_Standard
#endif
#endif
  
    ! --- in/out ---------------------------------
    
    character(len=*), intent(in)      ::  spec_from
    character(len=*), intent(out)     ::  spec_to
    integer, intent(out)              ::  status
    
    ! --- const ----------------------------------

    character(len=*), parameter ::  rname = mname//'/UDUnits_Standard'
    
    ! --- local ----------------------------------

    ! --- begin ----------------------------------

#ifdef with_udunits
    call UDUnits1_Standard( spec_from, spec_to, status )
    IF_NOT_OK_RETURN(status=1)
#else
#ifdef with_udunits2
    call UDUnits2_Standard( spec_from, spec_to, status )
    IF_NOT_OK_RETURN(status=1)
#else
    write (error_message,'("UDUnits requires definition of either with_udunits or with_udunits2 macro")')
    status=1; return
#endif
#endif
    
    ! ok
    status = 0
    
  end subroutine UDUnits_Standard
  
  
  ! ***
  
  
  subroutine UDUnits_ConversionFactor( spec_from, spec_to, factor, status )

#ifdef with_udunits
    use UDUnits1, only : UDUnits1_ConversionFactor
#else
#ifdef with_udunits2
    use UDUnits2, only : UDUnits2_ConversionFactor
#endif
#endif
  
    ! --- in/out ---------------------------------
    
    character(len=*), intent(in)      ::  spec_from
    character(len=*), intent(in)      ::  spec_to
    real(8), intent(out)              ::  factor
    integer, intent(out)              ::  status
    
    ! --- const ----------------------------------

    character(len=*), parameter ::  rname = mname//'/UDUnits_ConversionFactor'
    
    ! --- local ----------------------------------

    ! --- begin ----------------------------------

#ifdef with_udunits
    call UDUnits1_ConversionFactor( spec_from, spec_to, factor, status )
    IF_NOT_OK_RETURN(status=1)
#else
#ifdef with_udunits2
    call UDUnits2_ConversionFactor( spec_from, spec_to, factor, status )
    IF_NOT_OK_RETURN(status=1)
#else
    write (error_message,'("UDUnits requires definition of either with_udunits or with_udunits2 macro")')
    status=1; return
#endif
#endif
    
    ! ok
    status = 0
    
  end subroutine UDUnits_ConversionFactor
  
  
end module UDUnits


